@@ -111,14 +111,17 @@ int rwhois_limit;
 /* Set to 0 to completely disable whois-servers.net service support */
 int enable_whoisservers;
 
+/* Timeout value for connect calls in seconds */
+int connect_timeout;
+
 void help_version(int onlyversion)
 {
   char *COPYRIGHT = _("Copyright (C) 1999-%d  Free Software Foundation, Inc.\n");
   char *LICENSE =  _("This program is free software with ABSOLUTELY NO WARRANTY; you may\n\
 redistribute it under the terms of the GNU General Public License.");
 
-  printf("%s %s %s, \n", PACKAGE, _("version"), VERSION);
-  printf(COPYRIGHT, 2005);
+  printf("%s %s %s, ", PACKAGE, _("version"), VERSION);
+  printf(COPYRIGHT, 2007);
   printf("%s\n\n", LICENSE);
   if (!onlyversion)
     {
