@@ -51,8 +51,18 @@
 #include <rwhois.h>
 #include <utils.h>
 
+#include <errno.h>
 #include <string.h>
 
+#ifdef HAVE_ICONV
+# include <iconv.h>
+# include <langinfo.h>
+#endif
+
+#ifdef LIBIDN
+# include <idna.h>
+#endif
+
 #ifdef ENABLE_NLS
 # include <libintl.h>
 # define _(s)  gettext(s)
@@ -65,12 +75,12 @@ int jwhois_query(struct s_whois_query *, char **);
 int
 main(int argc, char **argv)
 {
-  int optind, count = 0, ret;
-  char *qstring = NULL, *text, *cachestr;
+  int optind, count = 0, ret, rc = 0;
+  char *qstring = NULL, *text, *cachestr, *idn;
   struct s_whois_query wq;
 
-#ifdef ENABLE_NLS
   setlocale(LC_ALL, "");
+#ifdef ENABLE_NLS
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 #endif
@@ -84,6 +94,7 @@ main(int argc, char **argv)
   /* Parse command line arguments and initialize the cache */
   optind = parse_args(&argc, &argv);
   cache_init();
+  timeout_init();
 
   /* Parse remaining arguments and place them into the wq
      structure. */
@@ -106,7 +117,18 @@ main(int argc, char **argv)
       optind++;
     }
   qstring[strlen(qstring)-1] = '\0';
+#ifdef LIBIDN
+  rc = idna_to_ascii_lz(qstring, &idn, 0);
+  if (rc != IDNA_SUCCESS)
+    {
+      printf("[IDN encoding of '%s' failed with error code %d]\n", qstring, rc);
+      exit(1);
+    }
+  wq.query = strdup(idn);
+  free(idn);
+#else
   wq.query = qstring;
+#endif
 
   if (ghost)
     {
@@ -131,16 +153,17 @@ main(int argc, char **argv)
   text = NULL;
 
 #ifndef NOCACHE
+  cachestr = malloc(strlen(wq.query) + strlen(wq.host) + 1);
+  if (!cachestr)
+    {
+      printf("[%s]\n", _("Error allocating memory"));
+      exit(1);
+    }
+  snprintf(cachestr, strlen(wq.query) + strlen(wq.host) + 1, "%s:%s",
+           wq.host, wq.query);
+
   if (!forcelookup && cache) {
     if (verbose>1) printf("[Looking up entry in cache]\n");
-    cachestr = malloc(strlen(wq.query) + strlen(wq.host) + 1);
-    if (!cachestr)
-      {
-        printf("[%s]\n", _("Error allocating memory"));
-        exit(1);
-      }
-    snprintf(cachestr, strlen(wq.query) + strlen(wq.host) + 1, "%s:%s",
-             wq.host, wq.query);
     ret = cache_read(cachestr, &text);
     if (ret < 0)
       {
@@ -172,6 +195,68 @@ main(int argc, char **argv)
   exit(0);
 }
 
+/*
+ * Attempt to convert string if result encoding is specified in the config
+ * file.
+ * */
+static char *
+convert_charset(struct s_whois_query *wq, char *curdata)
+{
+#ifdef HAVE_ICONV
+  const char *charset;
+
+  charset = get_whois_server_option(wq->host, "answer-charset");
+  if (charset != NULL)
+    {
+      iconv_t cd;
+
+      cd = iconv_open(nl_langinfo(CODESET), charset);
+      if (cd != (iconv_t)-1)
+	{
+	  char *buf, *src;
+	  size_t src_left, dest_size, dest_pos, res;
+
+	  src = curdata;
+	  src_left = strlen(src);
+	  dest_size = src_left;
+	  buf = malloc(dest_size);
+	  if (buf == NULL)
+	    goto error;
+	  dest_pos = 0;
+
+	  res = 0;
+	  while (src_left != 0)
+	    {
+	      char *dest;
+	      size_t dest_left;
+	      
+	      dest = buf + dest_pos;
+	      dest_left = dest_size - dest_pos;
+	      res = iconv(cd, &src, &src_left, &dest, &dest_left);
+	      if (res == (size_t)-1 && errno != E2BIG)
+		goto error;
+	      dest_pos = dest - buf;
+	      dest_size *= 2;
+	      buf = realloc(buf, dest_size);
+	      if (buf == NULL)
+		goto error;
+	    }
+	  buf[dest_pos] = 0;
+	  
+	  iconv_close(cd);
+	  free (curdata);
+	  return buf;
+	  
+	error:
+	  free(buf);
+	  iconv_close(cd);
+	}
+    }
+#endif
+  (void)wq;
+  return curdata;
+}
+
 /*
  *  This is the routine that actually performs a query. It selects
  *  the method to use for the host and then calls the correct routine
@@ -182,7 +267,7 @@ main(int argc, char **argv)
 int
 jwhois_query(struct s_whois_query *wq, char **text)
 {
-  char *tmp, *tmp2, *oldquery;
+  char *tmp, *tmp2, *oldquery, *curdata;
   int ret;
 
   if (!display_redirections)
@@ -196,20 +281,21 @@ jwhois_query(struct s_whois_query *wq, char **text)
 
   tmp = (char *)get_whois_server_option(wq->host, "rwhois");
   tmp2 = (char *)get_whois_server_option(wq->host, "http");
+  curdata = NULL;
 
   if ( (tmp && 0 == strcasecmp(tmp, "true")) || rwhois )
     {
-      ret = rwhois_query(wq, text);
+      ret = rwhois_query(wq, &curdata);
     }
   else
     {
       if (tmp2 && 0 == strcasecmp(tmp2, "true"))
 	{
-	  ret = http_query(wq, text);
+	  ret = http_query(wq, &curdata);
 	}
       else
 	{
-	  ret = whois_query(wq, text);
+	  ret = whois_query(wq, &curdata);
 	}
     }
 
@@ -223,6 +309,20 @@ jwhois_query(struct s_whois_query *wq, char **text)
     {
       exit(1);
     }
+  if (curdata != NULL)
+    {
+      curdata = convert_charset(wq, curdata);
+      if (*text == NULL)
+	*text = curdata;
+      else
+	{
+	  *text = realloc(*text, strlen (*text) + strlen (curdata) + 1);
+	  if (*text == NULL)
+	    exit(1);
+	  strcat(*text, curdata);
+	  free(curdata);
+	}
+    }
   if (ret > 0)
     {
       return jwhois_query(wq, text);
